#ifndef GEOMETRIES
#define GEOMETRIES

// sd prefix means signed distance
// ud prefix means unsigned distance
// primitives and other basic geometries goes here
// TODO: arbitrary meshes stored in textures. i have no idea yet how to calculate distance.
// TODO: texture coordinate mapping for other primitives than the sphere as well

struct geometrydata
{
	float distance;
	float2 texCoord;
};

float sdSphere( float3 p, float radius )
{
	return length(p)-radius;
}

geometrydata sdSphereC (float3 p, float3 center, float radius, float power)
{
	float3 P = center-p;
	float L = pow(pow(P.x,power) + pow(P.y,power) + pow(P.z,power), 1/power);

	geometrydata outg = (geometrydata)0;
	outg.distance = L-radius;
	outg.texCoord.x = atan2(P.z, length(P.xy));
	outg.texCoord.y = atan2(P.y, P.x);
	return outg;
}

float sdBox ( float3 p, float3 center, float3 extents )
{
	float3  di = abs(center-p) - extents;
	float mc = max(di.x, max(di.y, di.z));
	return min(mc,length(max(di,0.0)));
}

float udBox( float3 p, float3 center, float3 extents )
{
	return length(max(abs(center-p)-extents,0.0));
}

float udRoundBox( float3 p, float3 center, float3 extents, float radius )
{
	return length(max(abs(center-p)-extents,0.0))-radius;
}

float sdTorus( float3 p, float3 center, float2 radius, float2 power )
{
	p = center -p;
	float L1 = pow(pow(p.x,power.x) + pow(p.z,power.x), 1/power.x);
	float2 q = float2(L1-radius.x, p.y);
	float L2 = pow(pow(q.x,power.y) + pow(q.y,power.y), 1/power.y);
	return L2-radius.y;
}

float sdCylinder( float3 p, float3 c, float power )
{
	float2 P = p.xz-c.xy;
	float L = pow(pow(P.x,power) + pow(P.y,power), 1/power);
	return L-c.z;
}

float sdCone( float3 p, float2 c )
{
    c = normalize(c);
    float q = length(p.xy);
    return dot(c,float2(q,p.z));
}

float sdPlane( float3 p, float4 n )
{
  n.xyz = normalize(n.xyz);
  return dot(p,n.xyz) + n.w;
}

#endif // GEOMETRIES