#ifndef MODIFIERS
#define MODIFIERS

// write RayPosition modifiers or distance operations here

//////////////////// Transforms ////////////////////

float3 Repeat( float3 p, float3 distance, bool3 repeatOnAxis)
{
    float3 P = p;
    if(repeatOnAxis.x) P.x = (abs(P.x)%distance.x-0.5*distance.x)*sign ( P.x );
    if(repeatOnAxis.y) P.y = (abs(P.y)%distance.y-0.5*distance.y)*sign ( P.y );
    if(repeatOnAxis.z) P.z = (abs(P.z)%distance.z-0.5*distance.z)*sign ( P.z );
    return P;
}

float3 Transform( float3 p, float4x4 InvM)
{

    return mul(InvM,p);
}

// read 4x4 matrix transform data from textures
// format:
/*           tr01:       tr02:    ... tr0N:
    datarow1 datacol1    datacol2     datacolN
             matrix      matrix       matrix     // 16 rows of matrix components in left-to-right, top-to-bottom order

             tr11:       tr12:    ... tr1N:
    datarow2 datacol1    datacol2     datacolN
             matrix      matrix       matrix
               .          .     .      .
               .          .       .    .
               .          .         .  .

             trN1:       trN2:    ... trNN:
    datarowN datacol1    datacol2     datacolN
             matrix      matrix       matrix    */
float3 TransformTx( float3 p, sampler data, int dataColCount, int dataRowCount, int index)
{
    float4x4 m = 0;
    float2 coord = 0;
    for(int i = 0; i<16; i++)
    {
        coord.x = index%dataColCount / dataColCount;
        coord.y = (floor(index/dataColCount)*16+i)/(dataRowCount*16);
        m[i] = tex2D(data, coord);
    }
    
    return mul(m,p);
}

float3 Twist( float3 p, float2 freq )
{
    float c = cos(freq.x*p.y);
    float s = sin(freq.y*p.y);
    float2x2 m = float2x2(c,-s,s,c);
    float3 q = float3(mul(m,p.xz),p.y);
    return q;
}

float3 CheapBend( float3 p, float2 freq )
{
    float c = cos(freq.x*p.y);
    float s = sin(freq.y*p.y);
    float2x2 m = float2x2(c,-s,s,c);
    float3 q = float3(mul(m,p.xy),p.z);
    return q;
}

//////////////////// Operations ////////////////////
// for material aware operations
struct opdata
{
    float distance;
    int material;
};

// union
opdata oU(float d1, int MatID1, float d2, int MatID2)
{
    opdata t = (opdata)0;
    t.distance = min(d1, d2);
    if (t.distance == d1) t.material = MatID1;
    else t.material = MatID2;
    return t;
}

// subtract
opdata oS(float d1, int MatID1, float d2, int MatID2)
{
    opdata t = (opdata)0;
    t.distance = max(d1,-d2);
    if (t.distance == d1) t.material = MatID1;
    else t.material = MatID2;
    return t;
}

// intersection
opdata oI(float d1, int MatID1, float d2, int MatID2)
{
    opdata t = (opdata)0;
    t.distance = max(d1,d2);
    if (t.distance == d1) t.material = MatID1;
    else t.material = MatID2;
    return t;
}

// float material blending would be logical here it's not possible yet in this system
float oD(float d1, float d2) {return d1+d2;} // displace
float oBlend(float d1, float d2) {return smoothstep(d1,d2,d1-d2);}

#endif // MODIFIERS