// TODO: nicer shadows

#include "distanceField.fx"

struct shaderdata
{
	float lightness;
	int operations;
};

float3 NormalAtPoint ( float3 P, float epsilon)
{
	float3 n = (float3)0;
	float3 offset = float3 ( EPS, 0, 0 );
	int materialID;
	
	n.x = distanceField ( P +offset.xyz, materialID ) -distanceField ( P -offset.xyz, materialID );
	n.y = distanceField ( P +offset.yxz, materialID ) -distanceField ( P -offset.yxz, materialID );
	n.z = distanceField ( P +offset.yzx, materialID ) -distanceField ( P -offset.yzx, materialID );
	
	return normalize ( n );
}

//////////////////// AO ////////////////////

float AOTotal <string uiname = "AO Total (Global)";> = 0.0f;
float AOWeigth <string uiname = "AO Weight (Global)";> = 0.5f;
const int AOi <string uiname = "AO iterations (Global)";> = 5;

shaderdata AmbientOcclusion ( float3 P, float3 Norm )
{
	float total = AOTotal;
	float weigth = AOWeigth;
	int materialID;
	
	for ( int i = 0; i < AOi; ++i )
	{
		float delta = ( i +1 ) *( i +1 ) *sEPS;
		total += weigth *( delta -distanceField ( P +Norm *delta, materialID ) );
		weigth *= 0.5f;
	}

	shaderdata outS = (shaderdata)0;
	outS.lightness = 1.0f -saturate ( total );
	outS.operations = AOi;
	
	return outS;
}

//////////////////// SSS ////////////////////

float SSSTotal <string uiname = "SSS Total (Global)";> = 0.0f;
float SSSWeigth <string uiname = "SSS Weight (Global)";> = 0.5f;
const int SSSi <string uiname = "SSS iterations (Global)";> = 5;

shaderdata SubSurfaceScattering ( float3 P, float3 LookDirection )
{
	float total = SSSTotal;
	float weigth = SSSWeigth;
	int materialID;
	
	for ( int i = 0; i < 5; ++i )
	{
		float delta = pow ( i +1, 2.5f ) *sEPS;
		total += -weigth *min ( 0, distanceField ( P +LookDirection *delta, materialID ) );
		weigth *= 0.5f;
	}
	shaderdata outS = (shaderdata)0;
	outS.lightness = saturate ( total );
	outS.operations = SSSi;
	
	return outS;
}

//////////////////// Shadows ////////////////////

const float ShadowsMinT <string uiname="Shadow Bias (Global)";> = 0;
const float ShadowsMaxT <string uiname="Shadow Range (Global)";> = 10;
float ShadowsK <string uiname="Shadow Softness (Global)";> = 8;

shaderdata Shadows ( float3 P, float3 dir2light )
{
    float res = 1.0;
    int materialID;
    int shadi = 0;
    shaderdata outS = (shaderdata)0;
	float t=ShadowsMinT;
    while(t < ShadowsMaxT)
    {
        float h = distanceField(P + dir2light*t, materialID);
        if( h<sEPS )
    	{
    		outS.lightness = 0;
    		outS.operations = shadi;
            return outS;
        }
        res = min( res, ShadowsK*h/t );
        t += h;
        shadi++;
    }
    outS.lightness = res;
    outS.operations = shadi;
    return outS;
}

//////////////////// Fog ////////////////////

const float4 fogColor1 : COLOR <string uiname="Fog Color 1";> = float4 ( 0.5f, 0.6f, 0.7f, 1.0f );
const float4 fogColor2 : COLOR <string uiname="Fog Color 2";> = float4 ( 1.0f, 0.9f, 0.7f, 1.0f );

float3 postFog ( float3 currentColor, float3 ray, float3 sunDirection )
{
	float rayLength = length ( ray );
	ray = ray /rayLength;
	
	float fogAmount = 1.0f -exp ( -rayLength *0.008f );
	float sunAmount = pow ( max ( dot ( ray, sunDirection ), 0.0f ), 8.0f );
	
	float3 fogColor = lerp ( fogColor1.rgb, fogColor2.rgb, sunAmount );
	return lerp ( currentColor, fogColor, fogAmount );
}